from .stub import Stub
from .calls import Call
from .doubles import Spy
from .doubles import TestDouble
from .interfaces import Observer
from .wrapper import AttributeWrapper

__all__ = ('Spy', 'Call', 'Stub', 'TestDouble', 'Observer', 'AttributeWrapper')
